	/* wf - word frequency */

%{
/***********************************************************************
For two views of solving the word-frequency problem, see:

@Article{Bentley:litprog4,
  author =      "Jon Bentley and Donald E. Knuth and Doug McIlroy",
  title =       "Programming Pearls\emdash{}A Literate Program",
  journal =     CACM,
  year =        "1986",
  volume =      "29",
  number =      "6",
  pages =       "471--483",
  month =       jun,
}
***********************************************************************/
#include <string.h>
#include <ctype.h>
#undef yywrap		/* in case we use flex instead of lex */

#if __STDC__ && !__GNUC__
#include <stdlib.h>
#endif

#ifndef EXIT_FAILURE
#define EXIT_FAILURE	1
#endif

#define MAXHASH 32749	/* must be a PRIME int */

typedef struct
{
    char*	word;
    int		word_length;
    long	count;
}
Hash_Table_Entry;

static	Hash_Table_Entry hash_table[MAXHASH];

int	h;

int	compare();
int	hash();
int	lookup();
void	make_lower();
void	update();
%}

%%
[A-Za-z_][A-Za-z_0-9]*	{
    			    make_lower();
			    update(yytext,yyleng);
			}

\n |
.		;		/* ignore everything else */
%%
#define NBITS 5			/* shift count in hash function */

int
compare(a,b)			/* order a,b by descending counts */
Hash_Table_Entry	*a;
Hash_Table_Entry	*b;
{
    if (a->count > b->count)
	return (-1);
    else if (a->count < b->count)
	return (1);
    else
    {
	if (a->word[0] > b->word[0])
	    return (1);
	else if (a->word[0] < b->word[0])
	    return (-1);
	else
	    return (strcmp(a->word,b->word));
    }
}

int
hash(name)		/* return hash index of name in 0..MAXHASH-1 */
char *name;
{
    register int h;
    register char *p;

    for (h = 0, p = name; *p; ++p)
	h = ((h << NBITS) ^ *p);
    if (h < 0)
	h = -h;
    if (h < 0)	/* in case of most negative two's complement integer */
	h = 0;
    return (h % MAXHASH);
}

#if DEBUG
static long nl1, nl2, nl3, nl4, nl5;
#endif

int
lookup(name,name_length)
char *name;
int name_length;
{
    register int h;
    register int n;

#if DEBUG
    nl1++;
#endif
    for (n = 0, h = hash(name); n < MAXHASH; ++n)
    {
	if (hash_table[h].word == (char*)NULL)
	{
#if DEBUG
	    nl2++;
#endif
	    return (h);		/* found empty slot in hash_table[] */
	}
	else if ((name[0] == hash_table[h].word[0]) && /* OPTIMIZATION */
		 (hash_table[h].word_length == name_length) && /* 2nd OPT */
		(strcmp(name,hash_table[h].word) == 0))
	{
#if DEBUG
	    nl3++;
#endif
	    return (h);		/* found name in hash_table[] */
	}
#if DEBUG
	nl4++;
#endif
	h++;
	if (h >= MAXHASH)	/* check for wrap at end of table */
	    h = 0;
    }
#if DEBUG
    nl5++;
#endif
    return (-1);		/* table is full */
}

void
make_lower()
{
    register int n;

    for (n = 0; n < yyleng; ++n)
    {
	if (isupper(yytext[n]))
	    yytext[n] = tolower(yytext[n]);
    }
}

char*
Strdup(s)
char *s;
{
    char *p;
    p = (char*)malloc(strlen(s) + 1);
    if (p == (char*)NULL)
    {
	perror("Out of memory");
	exit (EXIT_FAILURE);
    }
    strcpy(p,s);
    return (p);
}

void
update(name,name_length)
char *name;
int name_length;
{
    int h = lookup(name,name_length);

    if (h >= 0)			/* then table is not yet full */
    {
	if (hash_table[h].word == (char*)NULL) /* add new name in table */
	{
	    hash_table[h].word = (char*)Strdup(name);
	    hash_table[h].word_length = name_length;
	}
	hash_table[h].count++;
    }
}

int
yywrap()
{
    int h;

#if DEBUG
    printf("lookup() counters: %ld %ld %ld %ld %ld\n",
	   nl1, nl2, nl3, nl4, nl5);
#endif

#if 1
    /* supply empty strings to simplify compare() code */
    for (h = 0; (h < MAXHASH); ++h)
	if (hash_table[h].word == (char*)NULL)
	    hash_table[h].word = "";

#if defined(__STDC__)
    qsort(hash_table, MAXHASH, sizeof(hash_table[0]),
	(int(*)(const void*,const void*))compare);
#else
    qsort(hash_table, MAXHASH, sizeof(hash_table[0]), compare);
#endif

    for (h = 0; (h < MAXHASH) && (hash_table[h].count > 0); ++h)
	printf("%ld: %s\n", hash_table[h].count, hash_table[h].word);
#else
    for (h = 0; (h < MAXHASH); ++h)
	if (hash_table[h].count > 0)
	    printf("%ld: %s\n", hash_table[h].count, hash_table[h].word);
#endif

    return (1);
}
